import unittest
from unittest.mock import MagicMock

from graafilohi import ExecutableDirectedMultiGraph


class TestExecutableDirectedMultiGraph(unittest.TestCase):
    def test_fully_executable_true(self):
        nodes = [1, 2, 3, 4]
        edges = [(1, 2), (2, 3), (3, 4)]

        g = ExecutableDirectedMultiGraph()
        for n in nodes:
            g.add_node(n, function=MagicMock())

        g.add_edges_from(edges)

        self.assertTrue(g.is_fully_executable())

    def test_fully_executable_false(self):
        edges = [(1, 2), (2, 3), (3, 4)]

        g = ExecutableDirectedMultiGraph()
        g.add_node(1)
        g.add_node(2, function=MagicMock())
        g.add_node(3, function=MagicMock())
        g.add_node(4, function=MagicMock())
        g.add_edges_from(edges)

        self.assertFalse(g.is_fully_executable())

    def test_node_function_not_callable(self):
        edges = [(1, 2), (2, 3), (3, 4)]

        g = ExecutableDirectedMultiGraph()
        g.add_node(1, function=MagicMock())
        g.add_node(2, function="I am not callable, but a string")
        g.add_node(3, function=MagicMock())
        g.add_node(4, function=MagicMock())
        g.add_edges_from(edges)

        self.assertFalse(g.is_fully_executable())

    def test_graph_has_cycle(self):
        nodes = [1, 2, 3, 4]
        edges = [(1, 2), (2, 3), (3, 4), (4, 3)]

        g = ExecutableDirectedMultiGraph()
        for n in nodes:
            g.add_node(n, function=MagicMock())

        g.add_edges_from(edges)

        self.assertFalse(g.is_fully_executable())

    def test_execution_order(self):
        nodes = [1, 2, 3, 4, 5, 6, 7, 8, 9, 0]
        edges = [(1, 2), (2, 4), (2, 5), (3, 4), (4, 6), (5, 6), (6, 7), (8, 7), (9, 7), (7, 0)]

        g = ExecutableDirectedMultiGraph()
        for n in nodes:
            g.add_node(n, function=MagicMock())

        g.add_edges_from(edges)

        order = g.get_execution_order(g)
        first = set(order[:4])
        second = order[4]
        third = set(order[5:7])
        last = set(order[7:])

        self.assertSetEqual(first, {1, 3, 8, 9})
        self.assertEqual(second, 2)
        self.assertSetEqual(third, {4, 5})
        self.assertSetEqual(last, {6, 7, 0})

    def test_execution(self):
        g = ExecutableDirectedMultiGraph()

        first_mock = MagicMock()
        second_mock = MagicMock()

        g.add_node(1, function=first_mock, function_arguments=[1, 2])
        g.add_node(2, function=second_mock)

        g.add_edge(1, 2)

        g.execute()

        first_mock.assert_called_once_with(1, 2, context={})
        second_mock.assert_called_once()

    def test_execution_context(self):
        g = ExecutableDirectedMultiGraph()

        first_mock = MagicMock(return_value=4)
        second_mock = MagicMock()

        g.add_node(1, function=first_mock, function_arguments=[1, 2])
        g.add_node(2, function=second_mock)

        g.add_edge(1, 2)

        g.execute()

        first_mock.assert_called_once_with(1, 2, context={})
        second_mock.assert_called_once_with(context={1: 4})
if __name__ == "__main__":
    unittest.main()


from setuptools import setup, find_packages

with open("VERSION") as vfile:
    version_line = vfile.readline()

with open("README.md", "r") as fh:
    long_description = fh.read()

version = version_line.strip()

setup(
    name="graafilohi",
    py_modules=["graafilohi"],
    include_package_data=True,
    package_data={
        "graafilohi": ["CHANGELOG.md", "VERSION", "README.md"]
    },
    version=version,
    license="MIT",
    description="Library for defining runnable pipelines as graphs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Aki Mäkinen",
    author_email="nenshou.sora@gmail.com",
    url="https://gitlab.com/blissfulreboot/python/graafilohi",
    keywords=[],
    install_requires=[
        "networkx",
        "matplotlib"
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6"
    ]
)



import networkx as nx
import matplotlib.pyplot as plt

from copy import deepcopy


DEFAULT_IMAGE_NAME = "dependency_graph.png"


class GraphNotFullyExecutableError(Exception):
    pass


class GraphNotExecutableError(Exception):
    pass


class ExecutableDirectedMultiGraph(nx.MultiDiGraph):
    def __init__(self, *args, **kwargs):
        super(ExecutableDirectedMultiGraph, self).__init__(*args, **kwargs)

    @staticmethod
    def search_root_nodes(graph):
        return [n for n in graph.nodes if graph.in_degree(n) == 0]

    @staticmethod
    def get_execution_order(graph):
        execution_order = []
        cloned_graph = deepcopy(graph)
        while len(cloned_graph.nodes) > 0:
            roots = ExecutableDirectedMultiGraph.search_root_nodes(cloned_graph)
            execution_order += roots
            for root_node in roots:
                cloned_graph.remove_node(root_node)
        return execution_order

    @staticmethod
    def get_node_colors(graph):
        color_map = []
        for node in graph.nodes:
            f = graph.nodes[node].get("function")
            if f is not None and callable(f):
                color_map.append("green")
            else:
                color_map.append("red")

        return color_map

    def is_fully_executable(self):
        for node in self.nodes:
            function = self.nodes[node].get("function")
            if not callable(function):
                return False

        try:
            nx.find_cycle(self)
            print("Graph contains a cycle")
            return False
        except nx.NetworkXNoCycle:
            pass

        return True

    def execute(self):
        if not self.is_fully_executable():
            raise GraphNotFullyExecutableError

        execution_order = self.get_execution_order(self)

        for node in execution_order:
            function_arguments = self.nodes[node].get("function_arguments", [])
            function_context = {p: self.nodes[p].get("return_value") for p in self.predecessors(node)}
            self.nodes[node]["return_value"] = \
                self.nodes[node].get("function")(*function_arguments, context=function_context)

    def partial_execute(self):
        raise NotImplementedError

    def simulate_execute(self):
        if not self.is_fully_executable():
            raise GraphNotFullyExecutableError

        execution_order = self.get_execution_order(self)

        print("The node functions will be executed in the following order: \n")
        print("\n".join(execution_order))

        print("Simulating the function calls:")
        for node in execution_order:
            function_arguments = self.nodes[node].get("function_arguments", [])
            function_name = self.nodes[node].get("function").__name__

            message = f"Calling function {function_name} "
            if len(function_arguments) > 0:
                message += f"with arguments {', '.join(function_arguments)}"
            print(message)

    def display_graph(self):
        color_map = self.get_node_colors(self)
        nx.draw(self, with_labels=True, node_color=color_map)
        plt.show()

    def save_graph(self, name=DEFAULT_IMAGE_NAME):
        color_map = self.get_node_colors(self)
        nx.draw(self, with_labels=True, node_color=color_map)
        plt.savefig(name)

    def save_and_display_graph(self, name=DEFAULT_IMAGE_NAME):
        color_map = self.get_node_colors(self)
        nx.draw(self, with_labels=True, node_color=color_map)
        plt.savefig(name)
        plt.show()
